#include "map.h"

#include "gfxengine.h"
#include "atlas.h"
#include "text.h"
#include <stdio.h>


MAP::MAP(const ROOM_DEF &def)
{
    myDef = &def;
    myWidth = def.size.x;
    myHeight = def.size.y;
    mySurface = false;

    mySquares = new SQUARE_NAMES[myWidth * myHeight];
    myMobCache = new MOB *[myWidth * myHeight];
    myItemCache = new ITEM *[myWidth * myHeight];
    myFlags = new u8[myWidth * myHeight];

    myName.reference(def.name);

    if (!myName.strcmp("world"))
    {
	mySurface = true;
    }

    myAtlas = 0;
    myMapId = -1;
    myLevel = def.level;

    memset(myMobCache, 0, myWidth * myHeight * sizeof(MOB *));
    memset(myItemCache, 0, myWidth * myHeight * sizeof(ITEM *));

    resetAll();
}

MAP::~MAP()
{
    for (int i = myItemList.entries(); i --> 0; )
    {
	ITEM		*it = myItemList(i);
	delete it;
    }
    for (int m = myMobList.entries(); m --> 0; )
    {
	MOB		*mob = myMobList(m);
	delete mob;
    }
    delete [] mySquares;
    delete [] myMobCache;
    delete [] myItemCache;
    delete [] myFlags;
}

void
MAP::resetAll()
{
    int		x, y;

    memset(mySquares, 0, myWidth * myHeight * sizeof(SQUARE_NAMES));
    memset(myFlags, 0, myWidth * myHeight * sizeof(u8));

    for (int mi = myMobList.entries(); mi --> 0; )
    {
	MOB	*m = myMobList(mi);
	if (!m->isAvatar())
	{
	    delete m;
	    // Do not do removeAt as that is done by the mob deletion!
	}
    }

    for (int mi = myItemList.entries(); mi --> 0; )
    {
	ITEM	*m = myItemList(mi);
	delete m;
	// Do not do removeAt as that is done by the mob deletion!
    }

    FORALL_XY(x, y)
    {
	mySquares[x + y * myWidth] = defn().squarelist[x + y * myWidth];
    }

    for (const PT2 *flag = defn().squareflaglist; flag && flag->data; flag++)
    {
	setFlag((FLAG_NAMES) flag->data, flag->x, flag->y, true);
    }

    // Set our flags from global flags
    FORALL_XY(x, y)
    {
	setFlag((FLAG_NAMES) defn().globalmapflags, x, y, true);
    }

    for (const PT2 *item = defn().itemlist; item && item->data; item++)
    {
	ITEM		*i = new ITEM((ITEM_NAMES) item->data);

	i->moveTo(buildPos(item->x, item->y));
    }

    for (const PT2 *mob = defn().moblist; mob && mob->data; mob++)
    {
	MOB		*m = new MOB((MOB_NAMES) mob->data);

	m->moveTo(buildPos(mob->x, mob->y));
    }

    myLevel = defn().level;
}

void
MAP::heartbeat(int phase)
{
    // Downwards as movement may delete mobs!
    for (int m = myMobList.entries(); m --> 0; )
    {
	if (myMobList(m)->heartbeat(phase))
	{
	    myMobList(m)->doAI();
	}
    }

    for (int i = myItemList.entries(); i --> 0; )
    {
	myItemList(i)->heartbeat(phase);
    }

    int		x, y;

    // Spread forest fires.
    // We do this first so we don't spread them on their
    // first run.
    FORALL_XY(x, y)
    {
	POS	p = buildPos(x, y);
	DIR	dir;
	POSLIST	setafire;		// Delay lighting to avoid
					// instant propagation as far
					// as possible.
	if (p.defn().burning)
	{
	    FORALL_DIR(dir)
	    {
		POS	g = p.delta(dir);

		if (g.defn().burn != SQUARE_NONE)
		{
		    // Potential to spread!
		    if (rand_chance(20))
			setafire.append(g);
		}
	    }

	    if (rand_chance(20))
	    {
		p.setTile((SQUARE_NAMES) p.defn().burned);
	    }
	}

	for (int i = 0; i < setafire.entries(); i++)
	{
	    setafire(i).setTile((SQUARE_NAMES) setafire(i).defn().burn);
	}
    }

    // Apply fire effects!
    FORALL_XY(x, y)
    {
	if (getFlag(FLAG_FIRE, x, y))
	{
	    if (getFlag(FLAG_FIREFRESH, x, y))
	    {
		setFlag(FLAG_FIREFRESH, x, y, false);
	    }
	    else
	    {
		// End the fire!
		setFlag(FLAG_FIRE, x, y, false);
		setFlag(FLAG_BURNED, x, y, true);

		if (glb_squaredefs[getTile(x, y)].burn != SQUARE_NONE)
		{
		    setTile(x, y, (SQUARE_NAMES)glb_squaredefs[getTile(x, y)].burn);
		}
		POS	p = buildPos(x, y);
		if (p.item() && p.item()->defn().burnable)
		{
		    // Oops:
		    delete p.item();
		}
	    }
	}

    }


    // Spread lava!
    FORALL_XY(x, y)
    {
	POS	p = buildPos(x, y);
	DIR	dir;
	POSLIST	setafire;		// Delay lighting to avoid
					// instant propagation as far
					// as possible.
	if (p.tile() == SQUARE_LAVA)
	{
	    FORALL_DIR(dir)
	    {
		POS	g = p.delta(dir);

		if (g.defn().lavarisk)
		{
		    // Potential to spread!
		    if (rand_chance(10))
			setafire.append(g);
		}
	    }

	    if (rand_chance(20))
	    {
		p.setTile(SQUARE_FROZEN_LAVA);
	    }
	}

	for (int i = 0; i < setafire.entries(); i++)
	{
	    setafire(i).convertToLava();
	}
    }
}

void
MAP::clearMessages()
{
    for (int m = 0; m < myMobList.entries(); m++)
    {
	myMobList(m)->clearMessages();
    }
}

MAP::iterator::iterator()
{
    myMap = 0;
    x = y = 0;
}

MAP::iterator::iterator(const MAP *map)
{
    myMap = map;
    x = y = 0;
}

void
MAP::iterator::advance()
{
    if (y >= myMap->height())
	return;

    x++;
    if (x >= myMap->width())
    {
	x = 0;
	y++;
    }
}

POS
MAP::iterator::pos() const
{
    return myMap->buildPos(x, y);
}

bool
MAP::iterator::done() const
{
    if (!myMap)
	return true;
    return y >= myMap->height();
}

MAP::iterator
MAP::begin() const
{
    MAP::iterator	it(this);

    return it;
}

MOB *
MAP::findMobByType(MOB_NAMES mob) const
{
    for (int m = 0; m < myMobList.entries(); m++)
    {
	if (!myMobList(m))
	    continue;

	if (myMobList(m)->getDefinition() == mob)
	    return myMobList(m);
    }
    return 0;
}

int
MAP::countTopItems(ITEM_NAMES item) const
{
    int		x, y;
    int		count = 0;
    FORALL_XY(x, y)
    {
	if (getItem(x, y) && getItem(x,y)->getDefinition() == item)
	{
	    count++;
	}
    }
    return count;
}

int
MAP::findMapLink(int xpos, int ypos) const
{
    if (xpos == 1)
	xpos = width()-1;
    if (ypos == 1)
	ypos = height()-1;

    for (int i = 0; defn().linklist[i].data >= 0; i++)
    {
	if ((ypos == -1 && defn().linklist[i].x == xpos) ||
	    (xpos == -1 && defn().linklist[i].y == ypos))
	{
	    // Check the link number
	    if (defn().linklist[i].data == 100)
	    {
		if (xpos == -1)
		    return defn().linklist[i].x;
		if (ypos == -1)
		    return defn().linklist[i].y;
	    }
	}
    }
    cerr << "No map link found" << endl;
    return 0;
}


void
MAP::makePosValid(POS &p)
{
    if (validPos(p.x(), p.y()))
	return;

    MAP		*adj;
    if (p.x() < 0)
    {
	if (defn().west)
	{
	    adj = atlas()->findMap(defn().west);
	    if (adj)
	    {
		int		ay = adj->findMapLink(1, -1);
		int		my = findMapLink(0, -1);
		p = adj->buildPos(p.x()+adj->width(), p.y() + ay - my);
		adj->makePosValid(p);
		return;
	    }
	}
    }
    if (p.x() >= width())
    {
	if (defn().east)
	{
	    adj = atlas()->findMap(defn().east);
	    if (adj)
	    {
		int		ay = adj->findMapLink(0, -1);
		int		my = findMapLink(1, -1);
		p = adj->buildPos(p.x()-width(), p.y() + ay - my);
		adj->makePosValid(p);
		return;
	    }
	}
    }
    if (p.y() < 0)
    {
	if (defn().north)
	{
	    adj = atlas()->findMap(defn().north);
	    if (adj)
	    {
		int		ax = adj->findMapLink(-1, 1);
		int		mx = findMapLink(-1, 0);
		p = adj->buildPos(p.x() + ax - mx, p.y()+adj->height());
		adj->makePosValid(p);
		return;
	    }
	}
    }
    if (p.y() >= height())
    {
	if (defn().south)
	{
	    adj = atlas()->findMap(defn().south);
	    if (adj)
	    {
		int		ax = adj->findMapLink(-1, 0);
		int		mx = findMapLink(-1, 1);
		p = adj->buildPos(p.x() + ax - mx, p.y()-height());
		adj->makePosValid(p);
		return;
	    }
	}
    }

    if (validPos(p.x(), p.y()))
	return;

    // Should check for portals here.
    p.myMap = 0;
}

void
MAP::setAtlas(ATLAS *atlas, int mapid)
{
    myAtlas = atlas;
    myMapId = mapid;
}

void
MAP::clear(SQUARE_NAMES square)
{
    for (int y = 0; y < height(); y++)
    {
	for (int x = 0; x < width(); x++)
	{
	    setTile(x, y, square);
	}
    }
}

bool
MAP::validPos(int x, int y) const
{
    if (x < 0 || x >= width())
	return false;
    if (y < 0 || y >= height())
	return false;

    return true;
}

SQUARE_NAMES
MAP::getTile(int x, int y) const
{
    if (!validPos(x, y))
	return SQUARE_NONE;

    return mySquares[x + y * width()];
}

ITEM *
MAP::getItem(int x, int y) const
{
    if (!validPos(x, y))
	return 0;

    return myItemCache[x + y * width()];
}

int
MAP::getItemList(int x, int y, ITEMLIST &list) const
{
    list.clear();

    POS		testpos = buildPos(x, y);

    for (int i = myItemList.entries(); i --> 0; )
    {
	if (myItemList(i)->pos() == testpos)
	    list.append(myItemList(i));
    }
    return list.entries();
}

void
MAP::setTile(int x, int y, SQUARE_NAMES square) 
{
    assert(validPos(x, y));

    mySquares[x + y * width()] = square;
}


void
MAP::setItem(int x, int y, ITEM *item) 
{
    assert(validPos(x, y));

    myItemCache[x + y * width()] = item;
}

void
MAP::getLinks(int x, int y, PTRLIST<int> &links) const
{
    for (int i = 0; defn().linklist[i].data >= 0; i++)
    {
	if (defn().linklist[i].x == x &&
	    defn().linklist[i].y == y)
	{
	    links.append(defn().linklist[i].data);
	}
    }
}

void
MAP::findLinks(int link, POSLIST &plist) const
{
    for (int i = 0; defn().linklist[i].data >= 0; i++)
    {
	if (defn().linklist[i].data == link)
	{
	    plist.append(buildPos(defn().linklist[i].x,
				  defn().linklist[i].y));
	}
    }
}

bool
MAP::getFlag(FLAG_NAMES flag, int x, int y) const
{
    if (!validPos(x, y))
	return SQUARE_NONE;

    return (myFlags[x + y * width()] & flag) ? true : false;
}

void
MAP::resetFlag(FLAG_NAMES flag, bool val) 
{
    int		x, y;

    if (val)
    {
	FORALL_XY(x, y)
	{
	    myFlags[x + y * width()] |= flag;
	}
    }
    else
    {
	FORALL_XY(x, y)
	{
	    myFlags[x + y * width()] &= ~((int)flag);
	}
    }
}

void
MAP::setFlag(FLAG_NAMES flag, int x, int y, bool val) 
{
    assert(validPos(x, y));

    if (val)
	myFlags[x + y * width()] |= flag;
    else
	myFlags[x + y * width()] &= ~((int)flag);
}

MOB *
MAP::getMob(int x, int y ) const
{
    if (!validPos(x, y))
	return 0;

    return myMobCache[x + y * width()];
}

void
MAP::setMob(int x, int y, MOB *mob)
{
    assert (validPos(x, y));

    myMobCache[x + y * width()] = mob;
}

void
MAP::registerMob(MOB *mob)
{
    myMobList.append(mob);
    assert(mob->pos().map() == this);

    setMob(mob->pos().myX, mob->pos().myY, mob);
}

void
MAP::deregisterMob(MOB *mob)
{
    myMobList.removePtr(mob);

    assert(getMob(mob->pos().myX, mob->pos().myY) == mob);
    if (getMob(mob->pos().myX, mob->pos().myY) == mob)
    {
	setMob(mob->pos().myX, mob->pos().myY, 0);
    }
}

void
MAP::updateMobCache(MOB *mob, POS oldpos, POS newpos)
{
    if (getMob(oldpos.myX, oldpos.myY) == mob)
	setMob(oldpos.myX, oldpos.myY, 0);

    setMob(newpos.myX, newpos.myY, mob);
}

void
MAP::registerItem(ITEM *item)
{
    // Corpses always go on the bottom.  But they must be on
    // top of other corpses or direction seems od.
    if (item->getDefinition() == ITEM_CORPSE)
    {
	int	corpsepos = 0;
	while (corpsepos < myItemList.entries())
	{
	    if (myItemList(corpsepos)->getDefinition() != ITEM_CORPSE)
	    {
		break;
	    }
	    corpsepos++;
	}

	myItemList.insert(corpsepos, item);
    }
    else
	myItemList.append(item);
    assert(item->pos().map() == this);

    // Overwrite with the top, with the exception corpses
    // always go below.
    if (item->getDefinition() != ITEM_CORPSE ||
	!getItem(item->pos().myX, item->pos().myY) ||
	getItem(item->pos().myX, item->pos().myY)->getDefinition() == ITEM_CORPSE
       )
    {
	setItem(item->pos().myX, item->pos().myY, item);
    }
}

void
MAP::deregisterItem(ITEM *item)
{
    myItemList.removePtr(item);

    // Note we aren't necesarily the same as the old item,
    // this will happen if we aren't on top.
    assert(getItem(item->pos().myX, item->pos().myY) == item);
    if (getItem(item->pos().myX, item->pos().myY) == item)
    {
	// Find topmost to replace.
	for (int i = myItemList.entries(); i --> 0; )
	{
	    if (myItemList(i)->pos() == item->pos())
	    {
		setItem(item->pos().myX, item->pos().myY, myItemList(i));
		break;
	    }
	}
	setItem(item->pos().myX, item->pos().myY, 0);
    }
}

void
MAP::updateItemCache(ITEM *item, POS oldpos, POS newpos)
{
    if (getItem(oldpos.myX, oldpos.myY) == item)
    {
	bool		found = false;
	// Find topmost to replace.
	for (int i = myItemList.entries(); i --> 0; )
	{
	    if (myItemList(i)->pos() == oldpos)
	    {
		setItem(oldpos.myX, oldpos.myY, myItemList(i));
		found = true;
		break;
	    }
	}
	if (!found)
	    setItem(oldpos.myX, oldpos.myY, 0);
    }

    setItem(newpos.myX, newpos.myY, item);
}

POS
MAP::buildPos(int x, int y) const
{
    POS		pos;

    pos.myX = x;
    pos.myY = y;
    pos.myMap = (MAP *)this;

    return pos;
}


static int glbDaystars = 0;

static void
settilestatus(int sx, int sy, POS pos)
{
    if (!pos.surface())
    {
	int		dark = pos.level() - glbDaystars;
	if (dark > 0)
	    gfx_settile_light(sx, sy, 255 - 32 * (pos.level()-1));
	else
	    gfx_settile_light(sx, sy, 255);
    }
    else
    {
	gfx_settile_light(sx, sy, 255);
    }
    gfx_settile_level(sx, sy, pos.level());
}

static void
renderTile(int sx, int sy, POS pos, bool reveal)
{
    if (sx < 0) return;
    if (sx >= SCR_WIDTH/4) return;
    if (sy < 0) return;
    if (sy >= SCR_HEIGHT/4) return;

    if (!pos.getFlag(FLAG_FOV))
    {
	gfx_settile(GFX_LAYER_TILES, sx, sy, TILE_BLEEDABLE_OPAQUE);
	gfx_settile(GFX_LAYER_ITEMS, sx, sy, TILE_NONE);
	gfx_settile(GFX_LAYER_MOBS, sx, sy, TILE_NONE);
	gfx_settile(GFX_LAYER_SHADOW, sx, sy, TILE_BLEEDABLE_TRANS);
	return;
    }

    settilestatus(sx, sy, pos);

    if (pos.getFlag(FLAG_SHADOW))
	gfx_settile(GFX_LAYER_SHADOW, sx, sy, TILE_SHADOW);
    else if (pos.getFlag(FLAG_FIRE))
	gfx_settile(GFX_LAYER_SHADOW, sx, sy, TILE_FIRE);
    else if (pos.getFlag(FLAG_DARK))
	gfx_settile(GFX_LAYER_SHADOW, sx, sy, TILE_SHADOW);
    else
	gfx_settile(GFX_LAYER_SHADOW, sx, sy, TILE_BLEEDABLE_TRANS);

    MOB		*mob;

    if (reveal && pos.defn().reveal != TILE_NONE)
	gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().reveal);
    else
    {
	if (glb_tiledefs[pos.defn().tile].wallset)
	{
	    // Select based on passability.
	    if (pos.south().passable() &&
		!pos.north().passable() &&
		!pos.east().passable() &&
		!pos.west().passable())
	    {
		gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, 1, 1);
		if (sy)
		{
		    gfx_settile(GFX_LAYER_DOUBLEHIGH, sx, sy-1, (TILE_NAMES) pos.defn().tile, 1, 0);
		    settilestatus(sx, sy-1, pos);
		}
	    }
	    else if (!pos.south().passable() &&
		!pos.north().passable() &&
		!pos.east().passable() &&
		!pos.west().passable() &&
		pos.south().east().passable())
	    {
		gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, 0, 1);
		if (sy)
		{
		    gfx_settile(GFX_LAYER_DOUBLEHIGH, sx, sy-1, (TILE_NAMES) pos.defn().tile, 0, 0);
		    settilestatus(sx, sy-1, pos);
		}
	    }
	    else if (!pos.south().passable() &&
		!pos.north().passable() &&
		!pos.east().passable() &&
		!pos.west().passable() &&
		pos.south().west().passable())
	    {
		gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, 2, 1);
		if (sy)
		{
		    gfx_settile(GFX_LAYER_DOUBLEHIGH, sx, sy-1, (TILE_NAMES) pos.defn().tile, 2, 0);
		    settilestatus(sx, sy-1, pos);
		}
	    }
	    else if (!pos.north().passable() &&
		!pos.south().passable() &&
		pos.east().passable() &&
		!pos.west().passable())
	    {
		gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, 0, 2);
	    }
	    else if (!pos.north().passable() &&
		!pos.south().passable() &&
		!pos.east().passable() &&
		pos.west().passable())
	    {
		gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, 2, 2);
	    }
	    else if (pos.north().passable() &&
		!pos.south().passable() &&
		!pos.east().passable() &&
		!pos.west().passable())
	    {
		gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, 1, 3);
	    }
	    else if (!pos.south().passable() &&
		!pos.north().passable() &&
		!pos.east().passable() &&
		!pos.west().passable() &&
		pos.north().east().passable())
	    {
		gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, 0, 3);
	    }
	    else if (!pos.south().passable() &&
		!pos.north().passable() &&
		!pos.east().passable() &&
		!pos.west().passable() &&
		pos.north().west().passable())
	    {
		gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, 2, 3);
	    }
	    else if (pos.south().passable() &&
		!pos.north().passable() &&
		!pos.east().passable() &&
		pos.west().passable())
	    {
		gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, 3, 2);
		if (sy)
		{
		    gfx_settile(GFX_LAYER_DOUBLEHIGH, sx, sy-1, (TILE_NAMES) pos.defn().tile, 3, 1);
		    settilestatus(sx, sy-1, pos);
		}
	    }
	    else if (pos.south().passable() &&
		!pos.north().passable() &&
		pos.east().passable() &&
		!pos.west().passable())
	    {
		gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, 4, 2);
		if (sy)
		{
		    gfx_settile(GFX_LAYER_DOUBLEHIGH, sx, sy-1, (TILE_NAMES) pos.defn().tile, 4, 1);
		    settilestatus(sx, sy-1, pos);
		}
	    }
	    else if (!pos.south().passable() &&
		pos.north().passable() &&
		!pos.east().passable() &&
		pos.west().passable())
	    {
		gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, 3, 0);
	    }
	    else if (!pos.south().passable() &&
		pos.north().passable() &&
		pos.east().passable() &&
		!pos.west().passable())
	    {
		gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, 4, 0);
	    }
	    else
	    {
		// Unknown pattern, default.
		gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, 1, 2);
	    }

	}
	else if (glb_tiledefs[pos.defn().tile].globalmap_x || glb_tiledefs[pos.defn().tile].globalmap_y)
	{
	    int		rx, ry;
	    rx = pos.x() % glb_tiledefs[pos.defn().tile].globalmap_x;
	    ry = pos.y() % glb_tiledefs[pos.defn().tile].globalmap_y;
	    gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, rx, ry);
	}
	else if (glb_tiledefs[pos.defn().tile].randset_x || glb_tiledefs[pos.defn().tile].randset_y)
	{
	    int		rx, ry;
	    rx = rand_wanginthash(rand_wanginthash(pos.x()) + pos.y());
	    rx += pos.y();
	    rx = rand_wanginthash(rx);
	    ry = rand_wanginthash(rx);
	    rx = abs(rx) % glb_tiledefs[pos.defn().tile].randset_x;
	    ry = abs(ry) % glb_tiledefs[pos.defn().tile].randset_y;
	    gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile, rx, ry);
	}
	else
	    gfx_settile(GFX_LAYER_TILES, sx, sy, (TILE_NAMES) pos.defn().tile);
    }

    ITEM		*item = pos.item();
    if (item)
    {
	int			toffx = 0, toffy = 0;
	TILE_NAMES		itemtile = item->tile(toffx, toffy);

	gfx_settile(GFX_LAYER_ITEMS, sx, sy, itemtile, toffx, toffy);
	if (sy && glb_tiledefs[itemtile].doublehigh)
	{
	    gfx_settile(GFX_LAYER_DOUBLEHIGH, sx, sy-1, itemtile, toffx, toffy-1);
	    settilestatus(sx, sy-1, pos);
	}
    }

    mob = pos.mob();
    if (mob)
    {
	if (!reveal && mob->intrinsic().get(INTRINSIC_INVIS))
	{
	    // You can see your self.
	    if (mob->isAvatar())
		gfx_settile(GFX_LAYER_MOBS, sx, sy, TILE_INVIS);
	}
	else
	{
	    if (mob->hp() == 0)
	    {
		// Pending death!
		int		deathy = glb_tiledefs[mob->defn().tile].fidget;
		if (!deathy)
		    deathy = 1;
		deathy += 2;
		gfx_settile(GFX_LAYER_MOBS, sx, sy, 
				(TILE_NAMES) mob->defn().tile,
				mob->getFacing().angle(), 
				deathy);
	    }
	    else if (mob->isAttacking())
	    {
		int		attacky = glb_tiledefs[mob->defn().tile].fidget;
		if (!attacky)
		    attacky = 1;
		switch (mob->getFacing().angle())
		{
		    case 0:		// North
			gfx_settile(GFX_LAYER_MOBS, sx, sy, 
					(TILE_NAMES) mob->defn().tile,
					2, attacky+1);
			if (sy)
			    gfx_settile(GFX_LAYER_DOUBLEHIGH, sx, sy-1, 
					(TILE_NAMES) mob->defn().tile,
					2, attacky);
			break;
		    case 1:		// West
			gfx_settile(GFX_LAYER_MOBS, sx, sy, 
					(TILE_NAMES) mob->defn().tile,
					1, attacky);
			if (sx)
			    gfx_settile(GFX_LAYER_DOUBLEHIGH, sx-1, sy, 
					(TILE_NAMES) mob->defn().tile,
					0, attacky);
			break;
		    case 2:		// South
			gfx_settile(GFX_LAYER_MOBS, sx, sy, 
					(TILE_NAMES) mob->defn().tile,
					3, attacky);
			if (sy+1 < SCR_HEIGHT/4)
			    gfx_settile(GFX_LAYER_DOUBLEHIGH, sx, sy+1, 
					(TILE_NAMES) mob->defn().tile,
					3, attacky+1);
			break;
		    case 3:		// East
			gfx_settile(GFX_LAYER_MOBS, sx, sy, 
					(TILE_NAMES) mob->defn().tile,
					0, attacky+1);
			if (sx+1 < SCR_WIDTH/4)
			    gfx_settile(GFX_LAYER_DOUBLEHIGH, sx+1, sy, 
					(TILE_NAMES) mob->defn().tile,
					1, attacky+1);
			break;
		}
	    }
	    else
	    {
		gfx_settile(GFX_LAYER_MOBS, sx, sy, (TILE_NAMES) mob->defn().tile,
				mob->getFacing().angle(), 0);
	    }
	}

	const char *msg = mob->getMessage();
	const char *emote = mob->getEmote();
	bool		hasemote = (emote && *emote);
	bool		hasmsg = (msg && *msg);
	if (hasemote || hasmsg)
	{
	    int		tx = sx * 4 + 3;
	    int		ty = sy * 4;
	    gfx_printstring(tx, ty-1, "/", hasemote ? ATTR_EMOTE : ATTR_SPEECH);
	    if (hasemote)
		gfx_printstring(tx+1-strlen(emote)/2, ty-2, emote, ATTR_EMOTE);
	    if (hasmsg)
		gfx_justifystring(tx+1, ty-2-hasemote, JUST_CENTER, JUST_BOTTOM, msg, ATTR_SPEECH);
	}
    }
    else
    {
	// We intentionaly put this on the mob layer as we
	// want it hidden by walking over it.
	if (pos.getFlag(FLAG_BURNED) && !pos.defn().burning)
	    gfx_settile(GFX_LAYER_MOBS, sx, sy, TILE_BURNED);
	else
	    gfx_settile(GFX_LAYER_MOBS, sx, sy, TILE_NONE);
    }
}

void
moveScreen(int &x, int &y, DIR dir, DIR face)
{
    switch ((dir.angle() + face.angle()) & 3)
    {
	case 0:
	    y--;
	    break;
	case 1:
	    x--;
	    break;
	case 2:
	    y++;
	    break;
	case 3:
	    x++;
	    break;
    }
}

void
MAP::render(POS center, DIR facedir, MOB *avatar)
{
    // Cache the daystar count once.
    glbDaystars = glbAtlas()->getDaystars();

    // Draw our intrisics
if (0)
{
    INTRINSIC_NAMES	intrinsic;
    int			statusx = 16;
    int			statusy = 2;
    FOREACH_INTRINSIC(intrinsic)
    {
	if (avatar->intrinsic().get(intrinsic))
	{
	    if (glb_intrinsicdefs[intrinsic].tile != TILE_NONE)
	    {
		gfx_settile(GFX_LAYER_ITEMS, statusx, statusy, (TILE_NAMES)glb_intrinsicdefs[intrinsic].over);
		gfx_settile(GFX_LAYER_TILES, statusx, statusy, (TILE_NAMES)glb_intrinsicdefs[intrinsic].tile);
		statusx++;
		if (statusx >= 19)
		{
		    statusy++;
		    statusx = 16;
		}
	    }
	}
    }

    while (statusy < 5)
    {
	gfx_settile(GFX_LAYER_ITEMS, statusx, statusy, TILE_NONE);
	gfx_settile(GFX_LAYER_TILES, statusx, statusy, TILE_NONE);
	statusx++;
	if (statusx >= 19)
	{
	    statusy++;
	    statusx = 16;
	}
    }
}


{
    //	Draw our items.
    int			itemx = 16;
    int			itemy = 2;
    int			numitem = avatar->numItems();

    for (int i = 0; i < numitem; i++)
    {
	const ITEM	*item = avatar->inventory(i);

	if (i == avatar->inventorySelection())
	    gfx_settile(GFX_LAYER_MOBS, itemx, itemy, TILE_SELECTOR);
	else
	    gfx_settile(GFX_LAYER_MOBS, itemx, itemy, TILE_NONE);

	int		toffx, toffy;
	TILE_NAMES itemtile = item->tile(toffx, toffy);
	gfx_settile(GFX_LAYER_ITEMS, itemx, itemy, itemtile, toffx, toffy);
	gfx_settile(GFX_LAYER_TILES, itemx, itemy, TILE_ITEMSLOT);
	itemx++;
	if (itemx >= 19)
	{
	    itemy++;
	    itemx = 16;
	}
    }

    while (itemy < 9)
    {
	gfx_settile(GFX_LAYER_MOBS, itemx, itemy, TILE_NONE);
	gfx_settile(GFX_LAYER_ITEMS, itemx, itemy, TILE_NONE);
	gfx_settile(GFX_LAYER_TILES, itemx, itemy, TILE_NONE);
	itemx++;
	if (itemx >= 19)
	{
	    itemy++;
	    itemx = 16;
	}
    }
}
#if 0
    // This doesn't look good :<
    gfx_printstring(60, 8, "Name:", ATTR_NORMAL);
    gfx_printstring(60, 9, text_lookupvariable("name"), ATTR_NORMAL);
#endif

    bool		reveal = false;
    if (avatar && avatar->intrinsic().get(INTRINSIC_REVEAL))
	reveal = true;
#if 1
    for (int y = 0; y < 15; y++)
    {
	for (int x = 0; x < 15; x++)
	{
	    for (int layer = 0; layer <= 4; layer++)
		gfx_settile(layer, x, y, TILE_NONE);
	}
    }
    DIR		dir;
    renderTile(7, 7, center, reveal);
    dir = DIR(0);
    FORALL_DIR(dir)
    {
	WAVEFRONT	front(center, dir);
	DIR		perp, backperp;
	int sx, sy;

	perp = dir.clock();
	backperp = dir.counterclock();
	sx = 7;
	sy = 7;

	for (int r = 1; r < SCR_RAD; r++)
	{
	    moveScreen(sx, sy, dir, facedir);
	    moveScreen(sx, sy, backperp, facedir);
	    int		wx = sx, wy = sy;
	    for (int i = 0; i < front.entries(); i++)
	    {
		renderTile(wx, wy, front(i), reveal);
		moveScreen(wx, wy, perp, facedir);
	    }
	    front.advance();
	}
    }
#else
    int		x, y;
    int		cx, cy;

    cx = center.x();
    cy = center.y();

    for (y = -1; y <= center.map()->height(); y++)
    for (x = -1; x <= center.map()->width(); x++)
    {
	int		sx, sy;
	
	sx = x - cx + 7;
	sy = y - cy + 7;

	POS		pos = center.map()->buildPos(x, y);

	renderTile(sx, sy, pos, reveal);
    }
#endif
}

u8 *
MAP::buildMiniMap() const
{
    u8	*rgb = new u8 [width() * height() * 3];
    int		x, y;

    FORALL_XY(x, y)
    {
	ATTR_NAMES		attr;

	ITEM		*item = getItem(x, y);
	attr = ATTR_NONE;
	if (item)
	    attr = (ATTR_NAMES) item->defn().minimap;
	if (attr == ATTR_NONE)
	    attr = (ATTR_NAMES)glb_squaredefs[mySquares[x + y * width()]].minimap;

	rgb[ (x + y * width()) * 3 + 0 ] = glb_attrdefs[attr].fg_b;
	rgb[ (x + y * width()) * 3 + 1 ] = glb_attrdefs[attr].fg_g;
	rgb[ (x + y * width()) * 3 + 2 ] = glb_attrdefs[attr].fg_r;
    }

    return rgb;
}

struct map_filterSquare
{
    map_filterSquare(SQUARE_NAMES square) { mySquare = square; }
    bool operator()(POS p) { return p.tile() == mySquare; }
    SQUARE_NAMES		mySquare;
};

struct map_filterItem
{
    map_filterItem(ITEM_NAMES item) { myItem = item; }
    bool operator()(POS p) { return p.item() && (p.item()->getDefinition() == myItem); }
    ITEM_NAMES		myItem;
};

struct map_filterPassable
{
    map_filterPassable() { }
    bool operator()(POS p) { return p.defn().passable; }
};

POS
MAP::findRandomSquare(SQUARE_NAMES square) const
{
    map_filterSquare filter(square);

    return findRandomPos(filter);
}

POS
MAP::findRandomItem(ITEM_NAMES item) const
{
    map_filterItem filter(item);

    return findRandomPos(filter);
}

POS
MAP::findRandomPassable() const
{
    map_filterPassable filter;

    return findRandomPos(filter);
}

int
MAP::validateLinks(bool quiet) const
{
    int		x, y;
    PTRLIST<int>		links;
    PTRLIST<int>		linearpos;
    POSLIST			dests;
    int				failed = 0;
    FORALL_XY(x, y)
    {
	POS	p = buildPos(x, y);

	switch (p.tile())
	{
	    case SQUARE_WALL_LEVER_UP:
	    case SQUARE_WALL_LEVER_DOWN:
		p.findLinkedSquares(dests);
		links.clear();
		getLinks(x, y, links);
		if (links.entries() != 1)
		{
		    if (!quiet)
			printf("Lever at %d,%d has %d out links!\n",
				x, y, links.entries());
		    failed++;
		}

		// Verify no duplicate destinations.
		linearpos.clear();
		for (int i = 0; i < dests.entries(); i++)
		    linearpos.append( dests(i).x() + (dests(i).y() <<16) );
		linearpos.stableSort();
		for (int i = 1; i < dests.entries(); i++)
		{
		    if (linearpos(i) == linearpos(i-1))
		    {
			int	lx, ly;
			lx = linearpos(i) & 65535;
			ly = linearpos(i) >> 16;
			if (!quiet)
			    printf("Lever at %d,%d controls duplicate at %d,%d\n", x, y, lx, ly);
			failed++;
		    }
		}

		// Must be at least one other thing controlled, which
		// means dests >= 2
		if (dests.entries() < 2)
		{
		    if (!quiet)
			printf("Lever at %d,%d controls nothing!\n", x, y);
		    failed++;
		}

		break;

	    case SQUARE_FLOOR_TELE:
	    case SQUARE_GROUND_TELE:
		p.findLinkedSquares(dests);
		if (dests.entries() != 2)
		{
		    if (!quiet)
			printf("Teleporter at %d,%d has %d destinations!\n",
				x, y, dests.entries());
		    failed++;
		}
		else
		{
		    // Make sure we have something other than ourselves!
		    if ( dests(0).x() == x && dests(0).y() == y &&
			 dests(1).x() == x && dests(1).y() == y )
		    {
			if (!quiet)
			    printf("Teleporter at %d,%d goes only to itself!\n",
				    x, y);
			failed++;
		    }
		}
		break;
	}
    }

    return failed;
}
